import '../mapper_exception.dart';
import '../mapper_utils.dart';
import 'mapper_base.dart';
import 'mapper_mixins.dart';
import 'simple_mapper.dart';

class PrimitiveMapper<T extends Object> extends MapperBase<T>
    with PrimitiveMethodsMixin<T> {
  const PrimitiveMapper([T Function(Object value)? decoder, this.exactType])
      : _decoder = decoder ?? _cast<T>;

  final T Function(Object value) _decoder;
  final Type? exactType;

  @override
  Type get type => exactType ?? super.type;
  @override
  bool isFor(dynamic v) =>
      exactType != null ? v.runtimeType == exactType : super.isFor(v);

  static T _cast<T>(v) => v as T;

  @override
  T decoder(Object value, DecodingContext context) {
    return _decoder(value);
  }

  @override
  Object encoder(T value, EncodingContext context) {
    return value;
  }
}

/// {@category Custom Mappers}
abstract class EnumMapper<T extends Enum> extends SimpleMapper<T> {
  const EnumMapper();
}

class DateTimeMapper extends SimpleMapper<DateTime> {
  const DateTimeMapper();

  @override
  DateTime decode(dynamic value) {
    if (value is String) {
      return DateTime.parse(value);
    } else if (value is num) {
      return DateTime.fromMillisecondsSinceEpoch(value.round());
    } else {
      throw MapperException.unexpectedType(value.runtimeType, 'String or num');
    }
  }

  @override
  String encode(DateTime self) {
    return self.toUtc().toIso8601String();
  }
}

typedef SerializableDecoder1<T, V> = T Function<A>(V, A Function(Object?));
typedef SerializableEncoder1<T> = Object Function(Object? Function(dynamic))
    Function(T);

typedef SerializableDecoder2<T, V> = T Function<A, B>(
    V, A Function(Object?), B Function(Object?));
typedef SerializableEncoder2<T> = Object Function(
        Object? Function(dynamic), Object? Function(dynamic))
    Function(T);

class SerializableMapper<T extends Object, V extends Object>
    extends MapperBase<T> with PrimitiveMethodsMixin<T> {
  late T Function(V value, DecodingContext context) _decoder;
  late Object Function(T value, EncodingContext context) _encoder;

  @override
  late Function typeFactory;

  @override
  bool includeTypeId<W>(v) => MapperBase.matchesStaticType<W>(v);

  SerializableMapper({
    required T Function(V) decode,
    required Object Function() Function(T) encode,
  })  : _decoder = ((v, c) => decode(v)),
        _encoder = ((v, c) => encode(v)()),
        typeFactory = ((f) => f<T>());

  SerializableMapper.arg1({
    required SerializableDecoder1<T, V> decode,
    required SerializableEncoder1<T> encode,
    required TypeFactory1 type,
  }) {
    _decoder =
        ((v, c) => c.call1(<A>(c) => decode<A>(v, c.container.fromValue<A>)));
    _encoder = ((v, c) =>
        c.call1(<A>(c) => encode(v)((o) => c.container.toValue<A>(o as A)))!);
    typeFactory = type;
  }

  SerializableMapper.arg2({
    required SerializableDecoder2<T, V> decode,
    required SerializableEncoder2<T> encode,
    required TypeFactory2 type,
  }) {
    _decoder = ((v, c) => c.call2(<A, B>(c) =>
        decode<A, B>(v, c.container.fromValue<A>, c.container.fromValue<B>)));
    _encoder = ((v, c) => c.call2(<A, B>(c) => encode(v)(
        (o) => c.container.toValue<A>(o as A),
        (o) => c.container.toValue<B>(o as B))));
    typeFactory = type;
  }

  @override
  T decoder(Object value, DecodingContext context) {
    return _decoder(value.checked<V>(), context);
  }

  @override
  Object encoder(T value, EncodingContext context) {
    return _encoder(value, context);
  }
}
